@@ -15,7 +15,7 @@
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
-from .base import Layout
+from libqtile.layout.base import Layout
 
 
 class _BspNode():
@@ -43,13 +43,16 @@ class _BspNode():
                 for c in child.clients():
                     yield c
 
-    def _shortest(self, l):
+    def _shortest(self, length):
         if len(self.children) == 0:
-            return self, l
-        else:
-            c0, l0 = self.children[0]._shortest(l + 1)
-            c1, l1 = self.children[1]._shortest(l + 1)
-            return (c1, l1) if l1 < l0 else (c0, l0)
+            return self, length
+
+        child0, length0 = self.children[0]._shortest(length + 1)
+        child1, length1 = self.children[1]._shortest(length + 1)
+
+        if length1 < length0:
+            return child1, length1
+        return child0, length0
 
     def get_shortest(self):
         return self._shortest(0)[0]
@@ -110,7 +113,7 @@ class Bsp(Layout):
     """This layout is inspired by bspwm, but it does not try to copy its
     features.
 
-    The first client occupies the entire srceen space.  When a new client
+    The first client occupies the entire screen space.  When a new client
     is created, the selected space is partitioned in 2 and the new client
     occupies one of those subspaces, leaving the old client with the other.
 
@@ -121,6 +124,8 @@ class Bsp(Layout):
     direction can be freely toggled.  All subspaces can be resized and
     clients can be shuffled around.
 
+    All clients are organized at the leaves of a full binary tree.
+
     An example key configuration is::
 
         Key([mod], "j", lazy.layout.down()),
@@ -143,11 +148,10 @@ class Bsp(Layout):
         Key([mod], "Return", lazy.layout.toggle_split()),
     """
     defaults = [
-        ("name", "bsp", "Name of this layout."),
         ("border_focus", "#881111", "Border colour for the focused window."),
         ("border_normal", "#220000", "Border colour for un-focused windows."),
         ("border_width", 2, "Border width."),
-        ("margin", 0, "Margin of the layout."),
+        ("margin", 0, "Margin of the layout (int or list of ints [N E S W])."),
         ("ratio", 1.6,
          "Width/height ratio that defines the partition direction."),
         ("grow_amount", 10, "Amount by which to grow a window/column."),
@@ -198,21 +202,21 @@ class Bsp(Layout):
             node.client = None
             self.current = self.root
 
-    def configure(self, client, screen):
-        self.root.calc_geom(screen.x, screen.y, screen.width,
-                            screen.height)
+    def configure(self, client, screen_rect):
+        self.root.calc_geom(screen_rect.x, screen_rect.y, screen_rect.width,
+                            screen_rect.height)
         node = self.get_node(client)
-        color = self.group.qtile.color_pixel(
-            self.border_focus if client.has_focus else self.border_normal)
+        color = self.border_focus if client.has_focus else self.border_normal
         border = 0 if node is self.root else self.border_width
-        client.place(
-            node.x,
-            node.y,
-            node.w - 2 * border,
-            node.h - 2 * border,
-            border,
-            color,
-            margin=self.margin)
+        if node is not None:
+            client.place(
+                node.x,
+                node.y,
+                node.w - 2 * border,
+                node.h - 2 * border,
+                border,
+                color,
+                margin=self.margin)
         client.unhide()
 
     def cmd_toggle_split(self):
